// use std::net::SocketAddr;

use std::{collections::HashMap, error::Error, fmt::Display};

use crate::utils::mime::mime_enum::Mime;

use super::{
    methods::Method,
    routes::{Data, Uri},
};

type HeaderMap = Vec<String>;
#[derive(Clone)]
pub struct Request<'a> {
    pub uri: Uri<'a>,
    pub headers: HeaderMap,
    pub method: Method,
    // pub connection: ConnectionMeta,
}

// struct ConnectionMeta {
//     remote: Option<SocketAddr>,
//     // certificates
// }

#[derive(Clone, Debug, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum MediaType {
    Json,
    Plain,
    Html,
}

#[derive(Debug)]
pub enum ParseErrors {
    NoData,
    BadData,
}

#[derive(Debug)]
pub struct ParseFormError {
    pub error: ParseErrors,
}

impl Display for ParseFormError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.error)
    }
}

impl Display for ParseErrors {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParseErrors::NoData => write!(f, "No Data at key"),
            ParseErrors::BadData => write!(f, "Bad Data at key"),
        }
    }
}

impl Error for ParseFormError {}

impl Request<'_> {
    pub fn can_have_body(&self) -> bool {
        match self.method {
            Method::Post | Method::Put | Method::Patch | Method::Delete => true,
            _ => false,
        }
    }
    pub fn mandatory_body(&self) -> bool {
        match self.method {
            Method::Post | Method::Put | Method::Patch => true,
            _ => false,
        }
    }
    // pub fn get_post_form_key<T: FromRequest>(&self, data: Data) -> T {}
    pub fn get_get_form_keys<'a>(
        &'a self,
        keys: &'a [&str],
    ) -> Result<HashMap<&str, &str>, ParseFormError> {
        let data = if let Some(val) = self.uri.split_once("?") {
            val
        } else {
            return Err(ParseFormError {
                error: ParseErrors::NoData,
            });
        };
        let data = data
            .1
            .split("&")
            .map(|kvp| kvp.split_once("="))
            .collect::<Vec<Option<(&str, &str)>>>();

        let mut values: HashMap<&str, &str> = HashMap::new();
        for kvp in data {
            let kvp = if let Some(kvp) = kvp {
                kvp
            } else {
                continue;
            };
            values.insert(kvp.0, kvp.1);
        }
        let mut response = HashMap::new();
        for key in keys {
            let entry = if let Some(val) = values.get_key_value(key) {
                val
            } else {
                return Err(ParseFormError {
                    error: ParseErrors::NoData,
                });
            };
            response.insert(*entry.0, *entry.1);
        }
        Ok(response)
    }
    pub fn get_post_text_form_key(
        &self,
        keys: &[&str],
        data: &Data,
    ) -> Result<Vec<String>, ParseFormError> {
        let mut post_type = self
            .headers
            .iter()
            .find(|header| header.contains("Content-Type: "))
            .unwrap()
            .to_string();

        post_type = post_type
            .strip_prefix("Content-Type: ")
            .unwrap()
            .to_string();

        let post_type: Vec<&str> = post_type.trim().split(';').collect();
        let mime_type = post_type[0].parse::<Mime>().unwrap();

        match mime_type {
            Mime::ApplicationXWwwFormUrlencoded => {
                let data = String::from_utf8(data.buffer.clone()).unwrap();
                let kvps = data
                    .split("&")
                    .map(|kvp| kvp.split_once("=").unwrap())
                    .collect::<HashMap<&str, &str>>();

                let mut result: Vec<String> = Vec::with_capacity(keys.len());
                for key in keys {
                    if let Some(val) = kvps.get(key) {
                        result.push(val.to_string());
                    } else {
                        return Err(ParseFormError {
                            error: ParseErrors::NoData,
                        });
                    }
                }
                Ok(result)
            }
            Mime::MultipartFormData => {
                let from_req = post_type[1..]
                    .iter()
                    .find(|val| val.contains("boundary="))
                    .unwrap()
                    .strip_prefix("boundary=")
                    .unwrap();
                let mut boundary = "--".as_bytes().to_vec();
                boundary.extend_from_slice(from_req.trim_matches('"').as_bytes());
                let mut end_boundary = boundary.clone();
                end_boundary.extend_from_slice(b"--");
                boundary.extend_from_slice(&[b'\r']);
                let parts = data
                    .buffer
                    .split(|byte| byte == &b'\n')
                    .collect::<Vec<&[u8]>>();

                let mut boundary_found = false;
                let mut current_key: Option<usize> = None;
                let mut result: Vec<Vec<u8>> = vec!["".into(); keys.len()];
                for part in parts {
                    if part == [] {
                        continue;
                    }
                    if part == end_boundary {
                        break;
                    }
                    if !boundary_found && part == boundary {
                        current_key = None;
                        boundary_found = true;
                        continue;
                    }
                    if part.starts_with(b"Content-Disposition: form-data; name=") {
                        let headers = part
                            .split(|byte| byte == &b';')
                            .filter(|header| !header.is_empty())
                            .collect::<Vec<_>>();
                        if headers.len() < 2 {
                            continue;
                        }
                        let name = headers[1].split(|byte| byte == &b'=').collect::<Vec<_>>();
                        if name.len() != 2 {
                            continue;
                        }
                        let mkey = String::from_utf8_lossy(name[1])
                            .to_string()
                            .trim_end()
                            .trim_matches('"')
                            .to_owned();
                        let mut index = 0;
                        for i in keys {
                            if *i == mkey {
                                current_key = Some(index);
                            }
                            index += 1;
                        }
                        boundary_found = false;
                    } else if let Some(key) = current_key {
                        result[key].extend_from_slice(part);
                        result[key].extend_from_slice(&[b'\n']);
                    }
                }
                Ok(result
                    .iter()
                    .map(|arr| String::from_utf8(arr.to_vec()).unwrap().trim().into())
                    .collect())
            }
            _ => Err(ParseFormError {
                error: ParseErrors::BadData,
            }),
        }
    }
}

#[cfg(test)]
mod test {
    use crate::handling::routes::Data;

    use super::Request;

    #[test]
    fn try_post_text() {
        let req = Request {
            uri: "",
            headers: vec!["Content-Type: multipart/form-data;boundary=\"boundary\"".to_string()],
            method: crate::handling::methods::Method::Post,
        };
        let data = Data {
            buffer: b"--boundary\r
Content-Disposition: form-data; name=\"field1\"\r
\r
value1\r
--boundary\r
Content-Disposition: form-data; name=\"field2\"; filename=\"example.txt\"\n\r
\r
value2\r
--boundary--"
                .to_vec(),
            is_complete: true,
        };
        assert_eq!(
            "value1",
            req.get_post_text_form_key(&["field1"], &data).unwrap()[0]
        );
        assert_eq!(
            "value2",
            req.get_post_text_form_key(&["field2"], &data).unwrap()[0]
        );
        assert_eq!(
            vec!["value1", "value2"],
            req.get_post_text_form_key(&["field1", "field2"], &data)
                .unwrap()
        )
    }
}
